#include "xml.h"

static bool isAlpha(int c) {return 'A'<=c && c<='Z' || 'a'<=c && c<='z' || c=='_';}
static bool isAlNum(int c) {return '0'<=c && c<='9' || c=='-' || c=='.' || c==':' || isAlpha(c);}

static const char entityNames[]="amp\0gt\0lt\0quot\0apos\0";
static const char entityList[]="&><\"'";

static const char elementNames[]=
	"catalog\0media\0directory\0file\0comment\0borrow\0mp3tag\0";
//	"datafile\0catlnk\0content\0archived\0";	// ungenutzt

extern volatile bool wantStop;

void fs::Entry::unlink() {
 prev->next=next;
 if (next) next->prev=prev;
 if (parent==this) parent=next;
}

const char*Xml::Stringbuf<>::removeEntities(bool trim) {
 char c,*dp=s;
 const char*sp=s;
 if (trim) while (sp!=s+slen) if ((byte)*sp<=' ') ++sp; else break;
 while(sp!=s+slen){
  c=*sp++;
  if (c=='&') {
   const char*a=sp;
   do{
    c=*sp++;	// Buchstaben und Ziffern suchen
   }while(isAlNum(c));
   const char*e=--sp;
   if (c==';') ++sp;
   for (const char*p=entityNames,*q=entityList;*p;q++) {
    int l=strlen(p);	// stets >0
    if (l==e-a && !memcmp(a,p,l)) {c=*q; goto found;}
    p+=l+1;
   }
   c='&'; sp=a;	// Unbekannte Entitt unverndert ausgeben
  }
found:
  *dp++=c;
 }
 if (trim) while (dp!=s) if ((byte)dp[-1]<=' ') --dp; else break;
 *dp=0;
 slen=dp-s;	// Neue Lnge
 return s;
}

static unsigned __int64 parseSize(const char*s) {
 unsigned ret=0;
 int nk=-1;
 char c;	// Parserschleife endet mit "K" oder "M" oder "G"
 while (c=*s++) {
  if (c<=' ') continue;
  if (c==',' || c=='.') nk=0;
  else if ('0'<=c && c<='9') {
   ret=ret*10+c-'0';
   if (nk>=0) ++nk;
  }else break;
 };
 size_t fak=1;
 switch (c) {
  case 'K':
  case 'k': fak=1000; break;	// dezimale Interpretation (stimmt das?)
  case 'M': fak=1000000; break;
  case 'G': fak=1000000000; break;
 }
 while (nk>0) {fak/=10; nk--;}
 return (unsigned __int64)(ret)*fak;
}

static unsigned parseDate(const char*s) {
 unsigned y,m,d,H,M,S;
 sscanf(s,"%u-%u-%u %u:%u:%u",&y,&m,&d,&H,&M,&S);
 return (y-1980)<<25	// 7 Bit
	| m<<21		// 4 Bit
	| d<<16		// 5 Bit
	| H<<11		// 5 Bit
	| M<<5		// 6 Bit
	| S>>1;		// 5 Bit
}

int Xml::parseAttr(fs::Entry*e, int c) {
 for(;;) {
  if (c<=' ') {
   if ((c=get())<0) return c;
   continue;	// skip whitespace
  }
  if (!isAlpha(c)) return c;
  Stringbuf<16>n;
  do{
   if (!(n+=c)) return c;
   if ((c=get())<0) return c;
  }while(isAlNum(c));
  Stringbuf<>v;
  if (c=='=') {
   int q=get();
   if (q!='"' && q!='\'') return c;	// must be in single or double quotes for XML, not for HTML
   for(;;) {
    if ((c=get())<0) return c;
    if (c==q) break;
    if (!(v+=c)) return c;
   }
   if ((c=get())<0) return c;		// Zeichen hinter Quote
  }
// einsortieren
  if (n=="name" || n=="title") {	// bei "catalog", "media", "directory" und "file"; bei "mp3tag"
   e->name=dup(v.removeEntities());
  }else if (n=="owner" || n=="artist") {	// bei "catalog" und "media"; bei "mp3tag"
   e->owner=dup(v.removeEntities());
  }else if (n=="size") {	// bei "file"
   e->size=parseSize(v);
  }else if (n=="type") {	// bei "media"
   byte m=fs::mHardDisk;
   static const char types[]="HardDisc\0CD\0DVD\0Diskette\0ZipDisk\0USB\0Cloud\0";
   for(const char*t=types;*t;t+=strlen(t)+1,++m) if (v==t) {e->medi=m; break;}
  }else if (n=="time") {	// bei "catalog", "media", "directory" und "file"
   e->ftLastMod=parseDate(v);
  }else if (n=="number") {	// bei "media"
   e->info = (byte)strtoul(v,0,0);
  }else if (n=="year") {	// bei "mp3tag"
   e->ftLastMod=strtoul(v,0,0);
  }
 }
}

bool Xml::parse(HANDLE h) {
 hStream=h;
 line=pos=idx=0;
 root.sub=0;
 buf=new char[BUFSIZE];
 bool ret=buf && readNext() && !(ec=innerParse());
 delete buf; buf=0;
 CloseHandle(hStream);
 return ret;
}

void Xml::runTo(const char*terminator) {
 Stringbuf<8>sb;
 size_t l=strlen(terminator);
 int c;
 do{
  if ((c=get())<0) return;
  sb+=c;
  if (sb.slen>l) memmove(sb.s,sb.s+1,--sb.slen);
 }while (sb!=terminator);
}

int Xml::innerParse() {
 fs::Entry*cur=&root;
 for (int c=get();c>=0;) if (c=='<') switch (c=get()) {
  case '!': {		// Comment und hnliches verschlucken
   Stringbuf<16>sb;
   for(;;) {
    if ((c=get())<0) return 25;
    if (c=='>') break;	// Kurzes unbekanntes Schlsselwort ignorieren
    if (sb=="DOCTYPE" && c<=' ') {runTo(">"); break;}
    else{
     if (!(sb+=c)) return 26;	// zu langes Schlsselwort
     if (sb=="--") {runTo("-->"); break;}
     else if (sb=="[CDATA[") {runTo("]]>"); break;}
    }
   }
   c=get();
  }break;
  case '/':
  if (cur==&root) return 16;
  cur=cur->parent;	// aufsteigen
  case '?':		// ProcessingInstruction verschlucken
  do{
   if ((c=get())<0) return 17;
  }while (c!='>');
  c=get();
  break;
  default:
  while ((byte)c<=' ') if ((c=get())<0) return 18;
  if (!isAlpha(c)) return 19;
  Stringbuf<16>n;
  do{
   if (!(n+=c)) return 20;
   if ((c=get())<0) return 21;
  }while (isAlNum(c));
  fs::Entry*e=newNode(cur);
  byte typ=eCatalog;
  e->type=eOther;		// irgendwas, nicht eRoot
  for (const char*en=elementNames;typ<=eMp3Tag;en+=strlen(en)+1,typ<<=1) {
   if (n==en) {e->type=typ; break;}
  }
  switch (c=parseAttr(e,c)) {
   case '/': if ((c=get())!='>') return 22; break;
   case '>': cur=e; break;	// absteigen
   default: return 23;		// Ungltiges Endezeichen
  }
  c=get();			// Zeichen hinter '>'
 }else{	// Text-Nodes
  if (wantStop) return -1;
  Stringbuf<>v;
  do{
   if (!(v+=c)) return 27;
   if ((c=get())<0) break;
  }while (c!='<');
  if (cur->type&(eComment|eBorrow)) {
   cur->name=dup(v.removeEntities(true));
  }
 }
 if (cur!=&root) return 24;
 return 0;	// okay
}

const Xml*Xml::locate(const fs::Entry*e) {
 while (e->parent) e=e->parent;	// aufsteigen bis root
 return (const Xml*)((const char*)e-offsetof(Xml,root));	// von root-Adresse auf Xml-Struktur schlieen
}

const char*Xml::dup(const char*s) const{
 size_t l=strlen(s)+1;
 char*p=(char*)HeapAlloc(hHeap,HEAP_GENERATE_EXCEPTIONS|HEAP_NO_SERIALIZE,l);
 memcpy(p,s,l);
 return p;
}

fs::Entry*Xml::newNode(fs::Entry*p) const{
 fs::Entry*e=(fs::Entry*)HeapAlloc(hHeap,HEAP_GENERATE_EXCEPTIONS|HEAP_NO_SERIALIZE|HEAP_ZERO_MEMORY,sizeof(fs::Entry));
 e->parent=p;
 if (p->sub) {
  e->prev=p->sub->prev;	// Das letzte Element wird zum vorletzten und <e> zum neuen letzten
  e->prev->next=e;	// Der Vorwrtszeiger des Vorletzten zeigt aufs letzte, also <e>
  p->sub->prev=e;	// Der Rckwrtszeiger des ersten zeigt auf den letzten, also <e>
 }else{
  p->sub=e;		// Das einzige Kind ist <e>
  e->prev=e;		// Der Rckwrtszeiger zeigt auf den letzten, also auf sich selbst
 }
 return e;
}

// Speicherblock lesen (kann bei get() passieren)
bool Xml::readNext() {
 idx=len=0;
 if (wantStop) return false;
 DWORD br;
 if (!ReadFile(hStream,buf,BUFSIZE,&br,0)) return false;
 len=br;
 return true;
}

bool Xml::write(HANDLE h) {
 hStream=h;
//TODO
 CloseHandle(hStream);
 return false;
}
