#pragma once
#define UNICODE
#define _UNICODE
#include <windows.h>
#include <windowsx.h>	// Edit_SetModify
#include <malloc.h>	// _alloca
#include <stdio.h>	// _snprintf
#define offsetof(s,m) ((size_t)&(((s*)0)->m))

#ifdef UNICODE
# define SA L"%S"	// String-Ansi
#else
# define SA "%s"
#endif

typedef unsigned char byte;	// wie Arduino

enum{
 eCatalog=1,
 eMedia=2,
 eDir=4,
 eFile=8,
 eComment=16,
 eBorrow=32,
 eMp3Tag=64,
 eOther=128,
};

/* Knotenverkettung am Beispiel: 3 Knoten, der mittlere hat kein Kindelement
   Whrend next am Ende 0 ist, zeigt prev auf das letzte Element.
   Von einem beliebigen Knoten (auer root) ist parent->sub->prev das letzte Element.
           ^
           +---<-------+---<-------+
+--------- | -->------ | -->------ | -+
|          |           |           |  |
| +------+ |  +------+ |  +------+ |  |
| | Node | |  | Node | |  | Node | |  |
| |parent|-+  |parent|-+  |parent|-+  |
| | next |--->| next |--->| next |>0  |
+-| prev |<---| prev |<---| prev |<---+
  | sub  |-+  | sub  |>0  | sub  |-+
  +------+ |  +------+    +------+ |
           v                       v
*/

namespace fs{

enum{
 mCatalog=0,
 mHardDisk=1,
 mCD=2,
 mDVD=3,
 mDiskette=4,
 mZipDisk=5,	// IOmega Zip100, Zip250, Jaz u.. Wechselmedien
 mFlash=6,	// SD-Karte, SD-Karte, CompactFlash
 mUsbMem=7,
 mNetwork=8,
 mCloud=9,
 mTape1=10,	// Magnetband mit 1 Spule
 mTape2=11,	// Magnetband in Kassette mit 2 Spulen
 mFilm=12,	// Filmrolle
 mVHS=13,	// VHS-Kassette u.. magnetischer Bildspeicher
};

struct Entry{
 Entry*sub,*parent,*prev,*next;
 byte type;		//catalog	media		directory	file		id3tag		comment		catlnk		archived	content		borrow
 byte attr;		//				FAT-attrib	FAT-attrib
 byte medi;		//		Medium-Typ
 byte info;		//		number						
 unsigned ftLastMod;	//		time		time		time		year
 unsigned __int64 size;	//						size		album
 const char*name;	//name		name		name		name		title		#text		name		type		#text		#text
 const char*owner;	//owner		owner						artist				location
 void unlink();
};

}

struct Xml {
 HANDLE hHeap;
 enum{BUFSIZE=1024*1024};
 unsigned line,pos,ec;	// Stelle des Fehlers bei parse(), 0-basiert
 bool parse(HANDLE h);	// linearisierte Version
 bool write(HANDLE h);
 fs::Entry root;	// nur <sub> wird benutzt, <parent> muss 0 bleiben
 static const Xml*locate(const fs::Entry*);
 const char*dup(const char*) const;	// Auf dem Heap
 fs::Entry*newNode(fs::Entry*) const;	// Auf dem Heap + Placement new
private:
 inline int get() {
  if (idx==len) readNext();	// Puffer ausgelesen? Nchsten Block lesen!
  if (idx==len) return -1;	// Puffer immer noch ausgelesen? EOF liefern
  int c=(byte)buf[idx++];	// Zeichen liefern
  if (c=='\n') {++line; pos=0;} else ++pos;	// Zeilen und Spalten mitzhlen
  return c;
 }
 HANDLE hStream;
 char*buf;
 unsigned idx,len;
 bool readNext();
 int innerParse();
 int parseAttr(fs::Entry*,int);
 void runTo(const char*);
 template<int MAX=3000> struct Stringbuf{
  size_t slen;
  char s[MAX];
  Stringbuf():slen(0) {}
//  void clear() {slen=0;}
  bool operator+=(char c) {if (slen>=sizeof s-1) return false; s[slen++]=c; return true;}
  bool operator==(const char*v) {s[slen]=0; return !strcmp(s,v);}
  bool operator!=(const char*v) {return !operator==(v);}
//  bool endsWith(const char*v) const {size_t lv=strlen(v); return slen>=lv && !memcmp(s+slen-lv,v,lv);}
  const char*removeEntities(bool=false);
  operator const char*() {s[slen]=0; return s;}
 };
};
